﻿using System;
using System.Collections.Generic;
using System.Text;
//using GroupLab.Networking;
//using GroupLab.SharedPhidgets;
//using GroupLab.SharedPhidgets.Utilities;
using System.Windows.Threading;
using System.Drawing;
using System.Windows;
using System.Windows.Controls;
using System.Threading;
//using System.Windows.Media;


namespace TTT
{
    class TTTArtefact
    {
        #region Constants

        /// <summary>
        /// Set the timer interval for the haptic loop
        /// </summary>
        public static int HAPTIC_LOOP_INTERVAL_IN_MILLISECONDS = 10;

        /// <summary>
        /// Set to true to inverse the height information of the servo: black = up, white = down.
        /// </summary>
        public static bool INVERSE_SERVO_HEIGHT = false;


        /// <summary>
        /// The maximum sensor value of the pressure sensor
        /// </summary>
        //public static double MAXIMUM_SENSOR_VALUE = 750.0;

        #endregion

        

        #region Variables

        /// <summary>
        /// Minimum servo value (upper position)
        /// </summary>
        private double servoMinimum = 0;

        /// <summary>
        /// Maximum servo value (lower position)
        /// </summary>
        private double servoMaximum = 180;

        /// <summary>
        /// Servo motor
        /// </summary>
        private Phidgets.AdvancedServoServo servoHeight = null;

        /// <summary>
        /// Servo motor
        /// </summary>
        private Phidgets.AdvancedServoServo servoBreak = null;

        /// <summary>
        /// Interfacekit: connected force sensor and slider
        /// </summary>
        private Phidgets.InterfaceKit interfaceKit = null;


        private int servoBreakPort = 0;
        private int servoHeightPort = 0;
        private int sensorPort;
        private double servoBreakMove;
        private double servoBreakStop;
        private double servoHeightMaximum;
        private double servoHeightMinimum; 

        HapticResponse hapticResponse = new HapticResponse();

        private DispatcherTimer timer = new DispatcherTimer();
        private MainWindow parent;

        private double amplitudeRange;
        private double hapticResponseRange;
        private double heightRange;
        private double currentHeight;
        private double currentOscillation;
        private double currentAmplitude;
        private double currentManualHeight;
        private double currentManualOscillation;
        private double currentManualAmplitude;
        private bool isBrakeOn;
        private HapticResponses currentHapticResponse;
        private int hapticLoopCounter = 0;
        private Color currentRatioColor = Color.White;
        private bool useDefaultRatios;


        /// <summary>
        /// Variables for the "picking pointer"
        /// </summary>
        private double pointerLength = 150.0;
        private double pointerOffsetLength = 40.0;
        private double pointerLengthTip = 155.0;
        private double pointerRotation = 180.0;
        private double currentAngle = 0.0;
        private double currentX = 0.0;
        private double currentY = 0.0;
        private double currentOffsetX = 0.0;
        private double currentOffsetY = 0.0;
        private double currentMarkerX = 0.0;
        private double currentMarkerY = 0.0;
        
        private string artefactID = "";
        private UIElement draggedElement = null;
        private TTTStatus currentStatus = TTTStatus.Idle;
        private System.Windows.Point lastDraggingPosition = new System.Windows.Point();


        private DemonstrationPreset currentPreset = DemonstrationPreset.Default;

        private bool bending = false;
        private System.Windows.Point bendingPoint = new System.Windows.Point();
        private System.Windows.Point previousPoint;
        private System.Windows.Media.Color paintColor = System.Windows.Media.Colors.Black;
        private System.Windows.Media.Color arrowColor = System.Windows.Media.Colors.Black;

        private int paymentDialogCounter = 1;
        private int paymentWaitTime = 0;
        private int paymentWaitCounter = 0;
        private bool paymentWait = false;



        private double maximumSensorValue = 750;

        #endregion


        #region Constructor


        /// <summary>
        /// Standard constructor
        /// </summary>
        /// <param name="servoHeightSerial"></param>
        /// <param name="servoHeightPort"></param>
        /// <param name="servoHeightMaximum"></param>
        /// <param name="servoHeightMinimum"></param>
        /// <param name="markerID"></param>
        /// <param name="sensorSerial"></param>
        /// <param name="sensorPort"></param>
        /// <param name="contactObserver"></param>
        /// <param name="parent"></param>
        public TTTArtefact(
            string servoHeightSerial, 
            int servoHeightPort,
            double servoHeightMaximum,
            double servoHeightMinimum, 
            string markerID, 
            string sensorSerial, 
            int sensorPort,
            MainWindow parent,
            System.Windows.Media.Color color,
            System.Windows.Media.Color arrowColor)
        {
            this.parent = parent;


            //this.artefactID = "TTT-ID:" + servoHeightSerial + "-" + servoHeightPort;
            this.artefactID = "TTT-ID:" + markerID;

            this.paintColor = color;
            this.arrowColor = arrowColor;

            this.servoHeight = this.parent.GetServoMotor(servoHeightPort);
            //this.servoHeight.Attach += new Phidgets.Events.AttachEventHandler(servoHeight_Attach);
            //this.servoHeight.FilterSerialNumbers.Add(servoHeightSerial);
            //int serial = int.Parse(servoHeightSerial);
            //this.servoHeight.open(serial);


            this.servoHeightMaximum = servoHeightMaximum;
            this.servoHeightMinimum = servoHeightMinimum;

            this.servoHeightPort = servoHeightPort;

            this.sensorPort = sensorPort;

            //this.interfaceKit = new Phidgets.InterfaceKit();
            //this.interfaceKit.FilterSerialNumbers.Add(sensorSerial);
            this.interfaceKit = this.parent.InterfaceKit;
            //this.interfaceKit.SensorChange += new Phidgets.Events.SensorChangeEventHandler(interfaceKit_SensorChange);

            

            // Register for move events
            this.parent.RegisterCallback(markerID, this);

            this.SetDefaultHapticRatios();

        }

        /*void interfaceKit_SensorChange(object sender, Phidgets.Events.SensorChangeEventArgs e)
        {
            throw new NotImplementedException();
        }*/

        private bool isManualControl = false;

        public bool ManualControl
        {
            get
            {
                return this.isManualControl;
            }
            set
            {
                this.isManualControl = value;
            }
        }


        private void SetDefaultHapticRatios()
        {
            // Set the balance between the different sorts of ratios             
            //SetHapticRatios(1, 0, 0); // only height
            //SetHapticRatios(0, 1, 0); // only haptic response
            //SetHapticRatios(0, 0, 1); // only oscillation
            SetHapticRatios(1, 1, 1); // balanced
        }

        public void SetMaximumSensorValue(double maximum)
        {
            this.maximumSensorValue = maximum;
        }

        public void SetPointerArrow(double pointerLength, double pointerOffsetLength)
        {
            this.pointerLength = pointerLength;
            this.pointerOffsetLength = pointerOffsetLength;
            this.pointerLengthTip = pointerLength + 5.0;
        }


        #endregion



        #region Haptic loop timer

        /// <summary>
        /// Calculates the ranges of each source of haptic information (height, oscillation, haptic response) according with the
        /// weights provided. Valid weight values go from 0 to infinite. Ranges are distributed proportionally according to weight.
        /// </summary>
        /// <param name="heightWeight">The weight of the height info source</param>
        /// <param name="hapticResponseWeight">The weight of the haptic response info source</param>
        /// <param name="oscillationWeight">The weight of the oscillation info source</param>
        public void SetHapticRatios(double heightWeight, double hapticResponseWeight, double oscillationWeight)
        {
            //only positive values are good. If negative values are provided, they are converted into positive.
            heightWeight = Math.Abs(heightWeight);
            hapticResponseWeight = Math.Abs(hapticResponseWeight);
            oscillationWeight = Math.Abs(oscillationWeight);

            amplitudeRange = oscillationWeight / (2 * (heightWeight + oscillationWeight + hapticResponseWeight));
            hapticResponseRange = hapticResponseWeight / (heightWeight + oscillationWeight + hapticResponseWeight);
            heightRange = heightWeight / (heightWeight + oscillationWeight + hapticResponseWeight);

            //System.Console.WriteLine("Ratio: amplRange-" + amplitudeRange + ", haptRange-" + hapticResponseRange + ", heightTange-" + heightRange);
        }

        public void SetDemonstrationPreset(DemonstrationPreset preset)
        {
            this.currentPreset = preset;
        }



        private double oscillationCycles = 0;
        private double oscillationCyclesMaximum = 0;
        private bool oscillationCyclesActive = false;

        public void SetOscillationPreset(OscillationPreset preset, int cycles)
        {
            this.oscillationCyclesActive = true;
            this.oscillationCycles = 0.0;
            this.oscillationCyclesMaximum = cycles * 2 * Math.PI;

            this.SetOscillationPreset(preset);

        }



        /// <summary>
        /// Select an oscillation preset
        /// </summary>
        /// <param name="preset"></param>
        public void SetOscillationPreset(OscillationPreset preset)
        {
            //this.oscillationCyclesActive = false;

            switch (preset)
            {
                case OscillationPreset.HighFreqLowAmplitude1:
                    this.currentManualAmplitude     = 0.1;
                    this.currentManualOscillation   = 1.0;
                    break;
                case OscillationPreset.HighFreqLowAmplitude2:
                    this.currentManualAmplitude = 0.16;
                    this.currentManualOscillation = 1.0;
                    break;
                case OscillationPreset.HighFreqMediumAmplitude:
                    this.currentManualAmplitude = 0.25;
                    this.currentManualOscillation = 1.0;
                    break;
                case OscillationPreset.HighFreqHighAmplitude:
                    this.currentManualAmplitude = 0.7;
                    this.currentManualOscillation = 1.0;
                    break;
                case OscillationPreset.MediumFreqHighAmplitude:
                    this.currentManualAmplitude = 0.7;
                    this.currentManualOscillation = 0.35;
                    break;
                case OscillationPreset.MediumFreqHighAmplitude2:
                    this.currentManualAmplitude = 0.5;
                    this.currentManualOscillation = 0.28;
                    break;
                case OscillationPreset.LowFrequencyLowAmplitude:
                    this.currentManualAmplitude = 0.28;
                    this.currentManualOscillation = 0.22;
                    break;
                case OscillationPreset.LowFrequencyHighAmplitude:
                    this.currentManualAmplitude = 0.7;
                    this.currentManualOscillation = 0.22;
                    break;
                case OscillationPreset.None:
                    this.currentManualAmplitude = 0.0;
                    this.currentManualOscillation = 0.0;
                    break;
                default:
                    break;
            }
        }


        /// <summary>
        /// Haptic loop cycle
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void hapticLoop_Cycle(object sender, EventArgs e)
        {
            if (this.currentPreset == DemonstrationPreset.Baskets)
            {
                this.hapticLoopMove();
            }
            else if (this.currentPreset == DemonstrationPreset.FloorPlanning)
            {
                this.hapticLoopFloorplanning();
            }
            else if (this.currentPreset == DemonstrationPreset.LayerPainting)
            {
                this.hapticLoopPaint();
            }
            else if (this.currentPreset == DemonstrationPreset.Art)
            {
                this.hapticLoopArt();
            }
            else if (this.currentPreset == DemonstrationPreset.Payment)
            {
                this.hapticLoopPayment();
            }
            else if (this.currentPreset == DemonstrationPreset.Pentimento)
            {
                this.hapticLoopPentimento();
            }
            else
            {
                this.hapticLoopDefault();
            }
        }




        private void hapticLoopDefault()
        {
            double currentSensorValue = this.GetNormalizedSensorValue();
            double currentOscillationSignalValue = currentAmplitude * Math.Sin(Math.PI * hapticLoopCounter * currentOscillation / 4.0 + Math.PI / 2);
            double currentHapticResponseValue = this.hapticResponse.GetHapticResponse(this.currentHapticResponse, currentSensorValue, 0);

            // Fusion of values
            double fusionValue = currentOscillationSignalValue * amplitudeRange +
                                 currentHapticResponseValue * hapticResponseRange +
                                 currentHeight * heightRange +
                                 amplitudeRange;

            this.SetHeight(fusionValue);

            // Brake
            if (isBrakeOn)
            {
                SetBreakStop();
            }
            else
            {
                SetBreakMove();
            }

            hapticLoopCounter++;
            if (hapticLoopCounter > 8 / currentOscillation) hapticLoopCounter = 0; 
        }



        private void hapticLoopPentimento()
        {
            double currentSensorValue = this.GetNormalizedSensorValue();
            double currentOscillationSignalValue = currentAmplitude * Math.Sin(Math.PI * hapticLoopCounter * currentOscillation / 4.0 + Math.PI / 2);
            double currentHapticResponseValue = this.hapticResponse.GetHapticResponse(HapticResponses.Soft, currentSensorValue, 0);

            // Fusion of values
            double fusionValue = currentHapticResponseValue * 0.5 +
                                 currentHeight * 0.3 + 0.2;

            this.SetHeight(fusionValue);

            //if (this.artefactID.Contains("258"))
            //{
                this.parent.ShowPentimento(this.artefactID, new System.Windows.Point(this.currentX, this.currentY), 1.0 - currentHapticResponseValue);
            //}

            hapticLoopCounter++;
            if (hapticLoopCounter > 8 / currentOscillation) hapticLoopCounter = 0;
        }




        private void hapticLoopPayment()
        {
            double currentSensorValue = this.GetNormalizedSensorValue();

            if (this.paymentWait)
            {
                this.paymentWaitCounter += 10;

                if (this.paymentWaitCounter > this.paymentWaitTime)
                {
                    this.paymentWaitCounter = 0;
                    this.paymentWait = false;
                    this.SelectNextPaymentDialog();

                }
                return;
            }



            if (this.parent.GetContainerIntersection(new System.Windows.Point(this.currentX, this.currentY)) == "none")
            {
                this.SetHeight(0.0);
            }
            else
            {
                double currentHapticResponseValue;

                if (paymentDialogCounter == 5)
                {
                    // Payment dialog

                    currentHapticResponseValue = this.hapticResponse.GetHapticResponse(HapticResponses.ButtonPressHard, currentSensorValue, 0);
                    this.SetHeight(currentHapticResponseValue);

                    if (currentHapticResponseValue == 0.0)
                    {
                        this.SetHeight(0.0);
                        this.SelectNextPaymentDialog();
                        this.PaymentWait(2000);                     
                        this.hapticResponse.ResetBuffer();
                    }
                }
                else
                {
                    // All other dialogs

                    currentHapticResponseValue = this.hapticResponse.GetHapticResponse(HapticResponses.ButtonPressVerySoft, currentSensorValue, 0);
                    this.SetHeight(currentHapticResponseValue);

                    if (currentHapticResponseValue < 0.3)
                    {
                        this.SetHeight(0.0);
                        this.SelectNextPaymentDialog();
                        this.PaymentWait(1000);
                        this.hapticResponse.ResetBuffer();
                    }

                }
            }
        }




        private void hapticLoopArt()
        {
            double currentSensorValue = this.GetNormalizedSensorValue();


            HapticResponses hr = HapticResponses.None;
            string id = this.parent.GetContainerIntersection(new System.Windows.Point(this.currentX, this.currentY));
            if (id != "none")
            {
                //System.Console.WriteLine("Intersect: " + id);
                this.SetOscillationPreset(OscillationPreset.MediumFreqHighAmplitude);
                hr = HapticResponses.ButtonPressSoft;

                if (this.hapticResponse.GetHapticResponse(hr, currentSensorValue, 0) == 0)
                {
                    this.parent.ShowInfoDialog(id, this.artefactID, new System.Windows.Point(currentX, currentY), currentAngle);
                }
                else
                {
                    this.parent.HideInfoDialog(this.artefactID);
                }

            }
            else
            {
                this.SetOscillationPreset(OscillationPreset.None);
                this.parent.HideInfoDialog(this.artefactID);
            }



            
            double currentHapticResponseValue = this.hapticResponse.GetHapticResponse(hr, currentSensorValue, 0);

            
            double currentOscillationSignalValue = currentManualAmplitude * Math.Sin(Math.PI * hapticLoopCounter * currentManualOscillation / 4.0 + Math.PI / 2);
            
            

            // Fusion of values
            double fusionValue = currentOscillationSignalValue * amplitudeRange +
                                 currentHapticResponseValue * hapticResponseRange +
                                 currentHeight * heightRange +
                                 amplitudeRange;

            this.SetHeight(fusionValue);

            hapticLoopCounter++;
            if (hapticLoopCounter > 8 / currentOscillation) hapticLoopCounter = 0;
        }




        private void hapticLoopPaint()
        {
            double currentSensorValue = this.GetNormalizedSensorValue();

            double threshold = 0.1;


            System.Windows.Point point = new System.Windows.Point(this.currentX, this.currentY);


            double currentHapticResponseValue = 
                this.hapticResponse.GetHapticResponse(HapticResponses.Paint, currentSensorValue, 0);

            this.SetHeight(currentHeight * 0.3 + currentHapticResponseValue * 0.5 + 0.2);

            if (currentSensorValue > threshold)
            {
                bool topLayer = false;
                if (currentHapticResponseValue == 0.5) topLayer = true;


                if (this.previousPoint.X == 0.0 && this.previousPoint.Y == 0.0)
                {
                    this.parent.Paint(point, point, this.paintColor, topLayer);
                }
                else
                {
                    this.parent.Paint(point, this.previousPoint, this.paintColor, topLayer);
                }
                this.previousPoint = point;
               
            }
            else
            {
                this.ResetDrawingPoint();
            }
        }




        private void hapticLoopMove()
        {
            double currentSensorValue = this.GetNormalizedSensorValue();

            //double threshold = 0.4;
            


            // Intersections
            string id = this.parent.GetContainerIntersection(new System.Windows.Point(this.currentX, this.currentY));
            if (id != "none")
            {
                //System.Console.WriteLine("Intersect: " + id);
            }



            //double currentOscillationSignalValue = currentAmplitude * Math.Sin(Math.PI * hapticLoopCounter * currentOscillation / 4.0 + Math.PI / 2);
            double currentHapticResponseValue = this.hapticResponse.GetHapticResponse(HapticResponses.TwoLayers, currentSensorValue, 0);

            if (currentHapticResponseValue <= 0.7)
            {
                if (this.currentStatus == TTTStatus.Idle)
                {
                    this.ExecuteHitTest(new System.Windows.Point(this.currentX, this.currentY));
                }
                else
                {
                    if (this.draggedElement != null)
                    {
                        double deltaX = this.lastDraggingPosition.X - this.currentX;
                        double deltaY = this.lastDraggingPosition.Y - this.currentY;
                        this.draggedElement.SetValue(Canvas.LeftProperty, (double)this.draggedElement.GetValue(Canvas.LeftProperty) - deltaX);
                        this.draggedElement.SetValue(Canvas.TopProperty, (double)this.draggedElement.GetValue(Canvas.TopProperty) - deltaY);
                        this.lastDraggingPosition = new System.Windows.Point(this.currentX, this.currentY);
                    }
                }
            }
            else
            {
                if (this.currentStatus == TTTStatus.Dragging)
                {
                    this.currentStatus = TTTStatus.Idle;
                    this.draggedElement = null;
                }
            }

            this.SetHeight(currentHapticResponseValue);

            // Brake
            if (isBrakeOn)
            {
                SetBreakStop();
            }
            else
            {
                SetBreakMove();
            }

            hapticLoopCounter++;
            if (hapticLoopCounter > 8 / currentOscillation) hapticLoopCounter = 0;  
        }




        public static double PRESSURE_LAYER_1 = 0.4;
        public static double PRESSURE_LAYER_2 = 0.7;


        private void hapticLoopFloorplanning()
        {
            double currentSensorValue = this.GetNormalizedSensorValue();
            bool activateBrake = false;

            // Intersections
            string id = this.parent.GetContainerIntersection(new System.Windows.Point(this.currentX, this.currentY));
            if (id.Contains("Brake"))
            {
                activateBrake = true;
            }

            double currentHapticResponseValue = 0.0;

            double HRratio = 0.7;
            double OSCratio = 0.3;

            double HRset = 0.0;
            double OSCset = 0.0;

            
            if (this.currentStatus == TTTStatus.Idle)
            {
                // No object is currently dragged
                // Check pressure status & check for intersections

                currentHapticResponseValue = this.hapticResponse.GetHapticResponse(HapticResponses.TwoLayers, currentSensorValue, 0);
                HRset = currentHapticResponseValue;
                //System.Console.WriteLine("HR: " + currentHapticResponseValue);

                // 0.7 for normal
                // 0.4 for really pressed

                UIElement element = this.GetHitTestResult();

                if (element != null)
                {
                    if (this.parent.IsDraggedObject((string)element.GetValue(Canvas.NameProperty)))
                    {
                        // Is already dragged by someone else
                        // Set oscillation


                        // When hard pressure, grab the object
                        if (currentHapticResponseValue <= PRESSURE_LAYER_1)
                        {
                            this.PickObject(element);
                        }
                    }
                    else
                    {
                        // The object is not dragged by someone else -> grab object
                        // when low pressure applied
                        if (currentHapticResponseValue <= PRESSURE_LAYER_2)
                        {
                            this.PickObject(element);
                        }
                    }
                }

            }
            else if (this.currentStatus == TTTStatus.Dragging)
            {
                // Currently dragging an object
                // Check for borders (if flag is true)

                currentHapticResponseValue = this.hapticResponse.GetHapticResponse(HapticResponses.TwoLayers, currentSensorValue, 0);
                HRset = currentHapticResponseValue;

                if (currentHapticResponseValue > PRESSURE_LAYER_2)
                {
                    this.ReleaseCurrentObject();
                }
                else if (currentHapticResponseValue == PRESSURE_LAYER_1)
                {
                    this.draggingOverride = true;
                    this.MoveDraggedObject(false);
                }
                else if (currentHapticResponseValue == PRESSURE_LAYER_2)
                {
                    this.MoveDraggedObject(true);
                }
                    /*
                else
                {
                    this.ReleaseCurrentObject();
                }*/

            }
            else if (this.currentStatus == TTTStatus.Override)
            {
                // This status occurs after the dragging of the object 
                // has been taken over by another HTP

                if (currentSensorValue == 0.0) this.currentStatus = TTTStatus.Idle;
                HRset = 1.0;
            }



            if (this.oscillationCyclesActive)
            {
                if (this.oscillationCyclesMaximum > this.oscillationCycles)
                {
                    OSCset =
                        this.currentManualAmplitude * Math.Sin(Math.PI *
                        hapticLoopCounter * this.currentManualOscillation / 4.0 + Math.PI / 2);

                    this.oscillationCycles += Math.PI * this.currentManualOscillation / 4.0;
                }
            }


            // Set the height
            this.SetHeight(HRset * HRratio + OSCset * OSCratio);

            // Activate brake if inside of blocked areas
            // or if bending is active
            if (activateBrake || this.bending) SetBreakStop();
            else SetBreakMove();



            hapticLoopCounter++;
            if (hapticLoopCounter > 8 / currentOscillation) hapticLoopCounter = 0;




            //
            // Previous code for picking/selecting
            //


            /*
            // Haptic response
            //double currentHapticResponseValue = this.hapticResponse.GetHapticResponse(HapticResponses.Breaky, currentSensorValue, 0);
            this.SetHeight(currentHapticResponseValue);
            //System.Console.WriteLine("HR: " + currentHapticResponseValue);

            // < 0.6 = level 1
            // < 0.3 = level 2

            
            bool pressureHard = false;
            if (currentSensorValue > 0.9)
            {
                pressureHard = true;
            }

            //double threshold = 0.4;
            //if (currentSensorValue > threshold)
            if(currentSensorValue > 0.3)
            {
                // picking
                if (this.currentStatus == TTTStatus.Idle)
                {
                    this.bending = false;
                    this.ExecuteHitTest(new System.Windows.Point(this.currentX, this.currentY));
                }
                // moving
                else
                {
                    if (this.draggedElement != null)
                    {
                        double deltaX = this.lastDraggingPosition.X - this.currentX;
                        double deltaY = this.lastDraggingPosition.Y - this.currentY;
                        double newLeft = (double)this.draggedElement.GetValue(Canvas.LeftProperty) - deltaX;
                        double newTop = (double)this.draggedElement.GetValue(Canvas.TopProperty) - deltaY;


                        if (!pressureHard)
                        {
                            // Check if moving object out of container
                            if (this.parent.IsInContainer(new Rect(
                                (double)this.draggedElement.GetValue(Canvas.LeftProperty),
                                (double)this.draggedElement.GetValue(Canvas.TopProperty),
                                (double)this.draggedElement.GetValue(Canvas.WidthProperty),
                                (double)this.draggedElement.GetValue(Canvas.HeightProperty))))
                            {
                                this.bending = false;
                                // move inside of container
                                this.draggedElement.SetValue(Canvas.LeftProperty, newLeft);
                                this.draggedElement.SetValue(Canvas.TopProperty, newTop);
                                this.lastDraggingPosition = new System.Windows.Point(this.currentX, this.currentY);
                            }
                            else
                            {
                                // bending
                                if (!bending)
                                {
                                    this.bending = true;
                                    this.bendingPoint = new System.Windows.Point(this.currentX, this.currentY);
                                }
                            }
                        }
                        else
                        {
                            this.bending = false;
                            // move inside of container
                            this.draggedElement.SetValue(Canvas.LeftProperty, newLeft);
                            this.draggedElement.SetValue(Canvas.TopProperty, newTop);
                            this.lastDraggingPosition = new System.Windows.Point(this.currentX, this.currentY);
                        }
                    }
                    else
                    {
                        this.bending = false;
                    }
                }
            }
            else
            {
                if (this.currentStatus == TTTStatus.Dragging)
                {
                    this.bending = false;
                    this.currentStatus = TTTStatus.Idle;
                    this.draggedElement = null;
                }
            }


            //double currentOscillationSignalValue = currentAmplitude * Math.Sin(Math.PI * hapticLoopCounter * currentOscillation / 4.0 + Math.PI / 2);


            */


        }



        private string draggingContainer = "";
        private bool draggingOverride = false;


        private void MoveDraggedObject(bool regionCheck)
        {
            if (this.draggedElement == null) return;

            double deltaX = this.lastDraggingPosition.X - this.currentX;
            double deltaY = this.lastDraggingPosition.Y - this.currentY;
            double newLeft = (double)this.draggedElement.GetValue(Canvas.LeftProperty) - deltaX;
            double newTop = (double)this.draggedElement.GetValue(Canvas.TopProperty) - deltaY;


            if ((regionCheck && (this.draggingContainer != "")) && !this.draggingOverride)
            {
                // Check if moving object outside of container
                if (this.parent.IsOutsideOfContainer(this.draggingContainer, new Rect(
                    newLeft,
                    newTop,
                    (double)this.draggedElement.GetValue(Canvas.WidthProperty),
                    (double)this.draggedElement.GetValue(Canvas.HeightProperty))))
                {

                    // bending
                    if (!bending)
                    {
                        this.bending = true;
                        this.bendingPoint = new System.Windows.Point(this.currentX, this.currentY);
                    }
                }
                else
                {
                    this.bending = false;
                    this.draggedElement.SetValue(Canvas.LeftProperty, newLeft);
                    this.draggedElement.SetValue(Canvas.TopProperty, newTop);
                    this.lastDraggingPosition = new System.Windows.Point(this.currentX, this.currentY);
                }

            }
            else
            {
                this.bending = false;
                this.draggedElement.SetValue(Canvas.LeftProperty, newLeft);
                this.draggedElement.SetValue(Canvas.TopProperty, newTop);
                this.lastDraggingPosition = new System.Windows.Point(this.currentX, this.currentY);
            }
        }



        public void ReleaseCurrentObject()
        {
            this.parent.NotifyActivity(this.artefactID);

            this.bending = false;
            this.draggingOverride = false;
            this.parent.UnsetDraggedObject(this.artefactID);
            this.currentStatus = TTTStatus.Idle;
            this.draggedElement = null;           
        }


        

        public void PickObject(UIElement element)
        {
            this.lastDraggingPosition = new System.Windows.Point(this.currentX, this.currentY);
            this.draggedElement = element;
            this.parent.SetDraggedObject(this.artefactID, (string)element.GetValue(Canvas.NameProperty));
            this.draggingContainer = this.parent.GetContainerIntersection(new Rect(
                        (double)this.draggedElement.GetValue(Canvas.LeftProperty),
                        (double)this.draggedElement.GetValue(Canvas.TopProperty),
                        (double)this.draggedElement.GetValue(Canvas.WidthProperty),
                        (double)this.draggedElement.GetValue(Canvas.HeightProperty)));
            this.currentStatus = TTTStatus.Dragging;
            
        }


        public void Override()
        {
            this.draggedElement = null;
            this.currentStatus = TTTStatus.Override;
        }



        private void ResetDrawingPoint()
        {
            this.previousPoint.X = 0.0;
            this.previousPoint.Y = 0.0;
        }



        private void SelectNextPaymentDialog()
        {
            //this.paymentDialogCounter++;
            //this.parent.InitializeLayersSimple(MainWindow.DEFAULT_DIRECTORY, "Screen" + this.paymentDialogCounter, "png");
            //Thread.Sleep(wait);

            this.paymentDialogCounter++;
            if (this.paymentDialogCounter > 8) this.paymentDialogCounter = 1;
            this.parent.InitializeLayersSimple(MainWindow.DEFAULT_DIRECTORY, "Screen" + this.paymentDialogCounter, "png");


            /*if (this.paymentDialogCounter == 4)
            {
                this.parent.InitializeLayersSimple(MainWindow.DEFAULT_DIRECTORY, "Screen4", "png");
                Thread.Sleep(1000);
                this.parent.InitializeLayersSimple(MainWindow.DEFAULT_DIRECTORY, "Screen5", "png");
                Thread.Sleep(1000);
                this.parent.InitializeLayersSimple(MainWindow.DEFAULT_DIRECTORY, "Screen1", "png");
                this.paymentDialogCounter = 1;
            }
            else
            {*/

            //}
        }

        private void PaymentWait(int time)
        {
            this.paymentWait = true;
            this.paymentWaitCounter = 0;
            this.paymentWaitTime = time;
        }





        private List<DependencyObject> hitResultsList = new List<DependencyObject>();

        public void ExecuteHitTest(System.Windows.Point pt)
        {
            // Retrieve the coordinate of the mouse position.
            //Point pt = e.GetPosition((UIElement)sender);

            // Clear the contents of the list used for hit test results.
            hitResultsList.Clear();

            // Set up a callback to receive the hit test result enumeration.
            System.Windows.Media.VisualTreeHelper.HitTest(this.parent.MainCanvas,
                null,
                new System.Windows.Media.HitTestResultCallback(HitTestResult),
                new System.Windows.Media.PointHitTestParameters(pt));

            // Perform actions on the hit test results list.
            if (hitResultsList.Count > 2)
            {

                //Console.WriteLine("Visuals hit: " + hitResultsList.Count + ", top: " + hitResultsList[1].GetType());
                
                if ((hitResultsList[2].GetType() == typeof(System.Windows.Shapes.Ellipse)) ||
                    (hitResultsList[2].GetType() == typeof(System.Windows.Controls.Image)))
                {
                    if (this.currentStatus == TTTStatus.Idle)
                    {
                        if (!this.parent.IsContainer((string)hitResultsList[2].GetValue(Canvas.NameProperty)))
                        {
                            this.currentStatus = TTTStatus.Dragging;
                            this.draggedElement = (UIElement)hitResultsList[2];
                            this.lastDraggingPosition = pt;
                        }
                    }
                    
                }
                //System.Windows.Shapes.Rectangle rect = new System.Windows.Shapes.Rectangle();
                //rect.RenderedGeometry.Bounds.Contains(

            }

        }



        public UIElement GetHitTestResult()
        {
            // Retrieve the coordinate of the mouse position.
            //Point pt = e.GetPosition((UIElement)sender);

            // Clear the contents of the list used for hit test results.
            hitResultsList.Clear();

            // Set up a callback to receive the hit test result enumeration.
            System.Windows.Media.VisualTreeHelper.HitTest(this.parent.MainCanvas,
                null,
                new System.Windows.Media.HitTestResultCallback(HitTestResult),
                new System.Windows.Media.PointHitTestParameters(new System.Windows.Point(this.currentX, this.currentY)));

            // Perform actions on the hit test results list.
            if (hitResultsList.Count > 2)
            {

                //Console.WriteLine("Visuals hit: " + hitResultsList.Count + ", top: " + hitResultsList[1].GetType());

                if ((hitResultsList[2].GetType() == typeof(System.Windows.Shapes.Ellipse)) ||
                    (hitResultsList[2].GetType() == typeof(System.Windows.Controls.Image)))
                {
                    return (UIElement)hitResultsList[2];

                        /*
                    if (this.currentStatus == TTTStatus.Idle)
                    {
                        if (!this.parent.IsContainer((string)hitResultsList[1].GetValue(Canvas.NameProperty)))
                        {
                            this.currentStatus = TTTStatus.Dragging;
                            this.draggedElement = (UIElement)hitResultsList[1];
                            this.lastDraggingPosition = pt;
                        }
                    }*/

                }
                //System.Windows.Shapes.Rectangle rect = new System.Windows.Shapes.Rectangle();
                //rect.RenderedGeometry.Bounds.Contains(

            }

            return null;

        }



        public System.Windows.Media.HitTestResultBehavior HitTestResult(System.Windows.Media.HitTestResult result)
        {
            // Add the hit test result to the list that will be processed after the enumeration.
            hitResultsList.Add(result.VisualHit);

            // Set the behavior to return visuals at all z-order levels.
            return System.Windows.Media.HitTestResultBehavior.Continue;
        }



        /// <summary>
        /// Get the normalized sensor value. Returns values in between 0.0 and 1.0
        /// </summary>
        /// <returns></returns>
        private double GetNormalizedSensorValue()
        {
            if (this.interfaceKit != null && this.interfaceKit.Attached)
            {
                double val = this.interfaceKit.sensors[this.sensorPort].Value / this.maximumSensorValue;
                if (val > 1.0) val = 1.0;
                else if (val < 0.0) val = 0.0;
                return val;
            }
            else return 0.0;
        }



        /// <summary>
        /// Start the haptic loop timer.
        /// The timer evaluates all the information from input layers, and determines
        /// the current artefact height and feedback behaviour.
        /// </summary>
        public void StartHapticLoop()
        {
            this.timer.Interval = new System.TimeSpan(0, 0, 0, 0, HAPTIC_LOOP_INTERVAL_IN_MILLISECONDS);
            this.timer.Tick += new EventHandler(hapticLoop_Cycle);
            this.timer.Start();
        }


        /// <summary>
        /// Stop the haptic loop timer.
        /// </summary>
        public void StopHapticLoop()
        {
            this.timer.Stop();
        }


        #endregion


        #region Servo Break

        public void InitializeBreak(
            string servoBreakSerial,
            int servoBreakPort,
            double servoBreakMove,
            double servoBreakStop)
        {
            this.servoBreak = this.parent.GetServoMotor(servoBreakPort);
            //this.servoBreak.FilterSerialNumbers.Add(servoBreakSerial);
            //int serial = int.Parse(servoBreakSerial);
            //this.servoBreak.open(serial);

            this.servoBreakPort = servoBreakPort;
            this.servoBreakMove = servoBreakMove;
            this.servoBreakStop = servoBreakStop;
        }


        public bool HasBreak
        {
            get
            {
                if (this.servoBreak != null) return true;
                else return false;
            }
        }

        public void SetBreakStop()
        {
            this.SetServoPositionBreak(this.servoBreakStop);
        }

        public void SetBreakMove()
        {
            this.SetServoPositionBreak(this.servoBreakMove);
        }

        private void SetServoPositionBreak(double position)
        {
            if (this.servoBreak != null && !this.isManualControl)
            {
                try
                {
                    this.servoBreak.Position = position;
                }
                catch (Phidgets.PhidgetException ex)
                {
                    
                }
            }
        }


        #endregion


        #region Servo Height

        private void SetServoPositionHeight(double position)
        {
            if (this.servoHeight != null && !this.isManualControl)
            {
                try
                {
                    this.servoHeight.Position = position;
                }
                catch (Phidgets.PhidgetException ex)
                {

                }
            }
        }

        private void SetHeight(double normalizedPosition)
        {
            //double setPosition = (normalizedPosition - this.servoHeightMinimum) / (this.servoHeightMaximum - this.servoHeightMinimum);
            if (!INVERSE_SERVO_HEIGHT)
            {
                normalizedPosition = 1.0 - normalizedPosition;
            }
            double setPosition = ((normalizedPosition * (this.servoHeightMaximum - this.servoHeightMinimum)) + this.servoHeightMinimum);

            if (setPosition > this.servoHeightMaximum) setPosition = this.servoHeightMaximum;
            else if (setPosition < this.servoHeightMinimum) setPosition = this.servoHeightMinimum;

            this.SetServoPositionHeight(setPosition);
        }




        #endregion





        /// <summary>
        /// Update the current position of the TTTArtefact.
        /// Gets the information of that position from all the haptic information layers.
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="angle"></param>
        public void UpdatePosition(double x, double y, double angle)
        {
            this.currentMarkerX = x;
            this.currentMarkerY = y;
            this.currentAngle = angle - Math.PI;// -this.pointerRotation;

            //this.currentX = this.currentMarkerX + Math.Cos(this.DegreeToRadian(this.currentAngle)) * this.pointerLength;
            //this.currentY = this.currentMarkerY + Math.Sin(this.DegreeToRadian(this.currentAngle)) * this.pointerLength;

            //this.currentOffsetX = this.currentMarkerX + Math.Cos(this.DegreeToRadian(this.currentAngle)) * this.pointerOffsetLength;
            //this.currentOffsetY = this.currentMarkerY + Math.Sin(this.DegreeToRadian(this.currentAngle)) * this.pointerOffsetLength;

            this.currentX = this.currentMarkerX + Math.Cos(this.currentAngle) * this.pointerLengthTip;
            this.currentY = this.currentMarkerY + Math.Sin(this.currentAngle) * this.pointerLengthTip;

            double drawCurrentX = this.currentMarkerX + Math.Cos(this.currentAngle) * this.pointerLength;
            double drawCurrentY = this.currentMarkerY + Math.Sin(this.currentAngle) * this.pointerLength;

            this.currentOffsetX = this.currentMarkerX + Math.Cos(this.currentAngle) * this.pointerOffsetLength;
            this.currentOffsetY = this.currentMarkerY + Math.Sin(this.currentAngle) * this.pointerOffsetLength;

            if (!this.bending)
            {
                this.parent.UpdatePointerArrow(this.artefactID, drawCurrentX, drawCurrentY, this.currentOffsetX, this.currentOffsetY, this.arrowColor);
            }
            else
            {
                this.parent.UpdatePointerArrow(this.artefactID, this.bendingPoint.X, this.bendingPoint.Y, this.currentOffsetX, this.currentOffsetY, this.arrowColor);
            }

            this.PickInformationFromLayers(this.currentX, this.currentY);        
        }



        internal void RemoveContact()
        {
            this.parent.RemovePointerArrow(this.artefactID);
        }

        private double DegreeToRadian(double angle)
        {
            return Math.PI * angle / 180.0;
        }


        public string GetArtefactID()
        {
            return this.artefactID;
        }



        private void PickInformationFromLayers(double x, double y)
        {
            this.currentHeight = (this.parent.GetPixelColor(MainWindow.LAYER_RELIEF, new System.Windows.Point(x, y))).B / 255.0;
            this.currentOscillation = (this.parent.GetPixelColor(MainWindow.LAYER_OSCILLATION, new System.Windows.Point(x, y))).B / 255.0;
            this.currentAmplitude = (this.parent.GetPixelColor(MainWindow.LAYER_AMPLITUDE, new System.Windows.Point(x, y))).B / 255.0;
            double tempBrake = (this.parent.GetPixelColor(MainWindow.LAYER_BRAKE, new System.Windows.Point(x, y))).B;
            this.isBrakeOn = tempBrake > 0;


            //
            // Set the haptic response function
            //
            Color hapticPixel = (this.parent.GetPixelColor(MainWindow.LAYER_HAPTICRESPONSE, new System.Windows.Point(x, y)));
            if (hapticPixel.Equals(HapticResponse.COLORCODE_HARD))
            {
                currentHapticResponse = HapticResponses.Hard;
            }
            else if (hapticPixel.Equals(HapticResponse.COLORCODE_SOFT))
            {
                currentHapticResponse = HapticResponses.Soft;
            }
            else if (hapticPixel.Equals(HapticResponse.COLORCODE_BREAKY))
            {
                currentHapticResponse = HapticResponses.Breaky;
            }
            else if (hapticPixel.Equals(HapticResponse.COLORCODE_VERYSOFT))
            {
                currentHapticResponse = HapticResponses.VerySoft;
            }
            else
            {
                currentHapticResponse = HapticResponses.None;
            }


            //
            // Set the haptic ratios
            //
            Color ratioColor = (this.parent.GetPixelColor(MainWindow.LAYER_RATIO, new System.Windows.Point(x, y)));

            if ((this.currentRatioColor == null) || (!this.currentRatioColor.Equals(ratioColor)))
            {
                this.currentRatioColor = ratioColor;

                if (this.currentRatioColor.Equals(Color.FromArgb(0, 0, 0)) || this.currentRatioColor.Equals(Color.FromArgb(0, 0, 0, 0)))
                {
                    this.SetDefaultHapticRatios();
                }
                else
                {

                    this.SetHapticRatios((double)this.currentRatioColor.R, (double)this.currentRatioColor.G, (double)this.currentRatioColor.B);
                }
            }
        }





        internal void NotifyActivity()
        {
            this.SetOscillationPreset(OscillationPreset.MediumFreqHighAmplitude2, 1);
        }
    }



    public enum TTTStatus
    {
        Idle, 
        Dragging,
        Override
    }


    public enum OscillationPreset
    {
        HighFreqLowAmplitude1,
        HighFreqLowAmplitude2,
        HighFreqMediumAmplitude,
        HighFreqHighAmplitude,
        MediumFreqHighAmplitude,
        MediumFreqHighAmplitude2,
        LowFrequencyLowAmplitude,
        LowFrequencyHighAmplitude,
        None
    }
}
